from typing import Callable

from optimus.helpers.types import *


class BaseSet():
    """Base class for all set implementations"""

    def __init__(self, root: 'DataFrameType'):
        self.root = root

    def __call__(self, cols=None, value=None, default=None, args=[], where=None):
        return self.root.cols.set(cols=cols, value_func=value, args=args, where=where, default=default)

    def all(self, cols, value=None) -> 'DataFrameType':
        return self.__call__(cols, value)

    def _mask(self, cols, true_value=None, default_value=None, args=[], func: Callable = None, **kwargs) -> 'DataFrameType':

        df = self.root

        if cols in df.cols.names():
            input = df[cols]
            input_cols = cols
        elif isinstance(true_value, (self.root.__class__, )):
            input = true_value
            input_cols = input.cols.names()[0]
        else:
            df[cols] = default_value or true_value
            input = df[cols]
            input_cols = cols

        # uses name instead of the function to use it correctly with 'input' instead of with the whole dataframe
        mask = getattr(input.mask, func.__name__)(input_cols, **kwargs)
        return self.__call__(cols, value=true_value, default=default_value, args=args, where=mask)

    # Types

    def str(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.str)

    def int(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.int)

    def float(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.float)

    def numeric(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.numeric)

    def email(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.email)

    def ip(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.ipv4)

    def url(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.url)

    def gender(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.gender)

    def boolean(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.boolean)

    def zip_code(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.zip_code)

    def credit_card_number(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.credit_card_number)

    def datetime(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.datetime)

    def object(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.object)

    def array(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.array)

    def phone_number(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.phone_number)

    def social_security_number(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.social_security_number)

    def http_code(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.http_code)

    # Other

    def null(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.null)

    def none(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.none)

    def nan(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.nan)

    def empty(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.empty)

    def greater_than(self, cols, true_value=None, default_value=None, args=[], value=None) -> 'DataFrameType':
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, func=self.root.mask.greater_than, value=value)

    def greater_than_equal(self, cols, true_value=None, default_value=None, args=[], value=None) -> 'DataFrameType':
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, func=self.root.mask.greater_than_equal, value=value)

    def less_than(self, cols, true_value=None, default_value=None, args=[], value=None) -> 'DataFrameType':
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, func=self.root.mask.less_than, value=value)

    def less_than_equal(self, cols, true_value=None, default_value=None, args=[], value=None) -> 'DataFrameType':
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, func=self.root.mask.less_than_equal, value=value)

    def between(self, cols, true_value=None, default_value=None, args=[], lower_bound=None, upper_bound=None, equal=True, bounds=None) -> 'DataFrameType':
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, func=self.root.mask.between, lower_bound=lower_bound, upper_bound=upper_bound, equal=equal, bounds=bounds)

    def equal(self, cols, true_value=None, default_value=None, args=[], value=None) -> 'DataFrameType':
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, func=self.root.mask.equal, value=value)

    def not_equal(self, cols, true_value=None, default_value=None, args=[], value=None) -> 'DataFrameType':
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, func=self.root.mask.not_equal, value=value)

    def missing(self, cols, value=None, default_value=None, args=[]) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.missing)

    def mismatch(self, cols, value=None, default_value=None, args=[], data_type=None) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.mismatch, data_type=data_type)

    def duplicated(self, cols, value=None, default_value=None, args=[], keep="first") -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.duplicated, keep=keep)

    def unique(self, cols, value=None, default_value=None, args=[], keep="first") -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.unique, keep=keep)

    def match(self, cols, value=None, default_value=None, args=[], regex=None, data_type=None) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.match_data_type, regex=regex, data_type=data_type)

    def match_regex(self, cols, value=None, default_value=None, args=[], regex=None) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.match_regex, regex=regex)

    def match_data_type(self, cols, value=None, default_value=None, args=[], data_type=None) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.match_data_type, data_type=data_type)

    def match_pattern(self, cols, value=None, default_value=None, args=[], pattern=None) -> 'DataFrameType':
        return self._mask(cols, true_value=value, default_value=default_value, args=args, func=self.root.mask.match_pattern, pattern=pattern)

    def value_in(self, cols, true_value=None, default_value=None, args=[], values=None) -> 'DataFrameType':
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, func=self.root.mask.value_in, values=values)

    def starts_with(self, cols, true_value=None, default_value=None, args=[], value=None) -> 'DataFrameType':
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, func=self.root.mask.starts_with, value=value)

    def ends_with(self, cols, true_value=None, default_value=None, args=[], value=None) -> 'DataFrameType':
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, func=self.root.mask.ends_with, value=value)

    def contains(self, cols, true_value=None, default_value=None, args=[], value=None) -> 'DataFrameType':
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, func=self.root.mask.contains, value=value)

    def expression(self, cols, true_value=None, default_value=None, args=[], where=None) -> 'DataFrameType':
        return self._mask(cols, true_value=true_value, default_value=default_value, args=args, func=self.root.mask.expression, where=where)



def dict_filter(dic, keys: list) -> list:
    """
    Get values from a dict  given a list of keys
    :param dic: dictionary to be filtered
    :param keys: list of keys to be used as filter
    :return:
    """
    return [(dic[i]) for i in keys if i in list(dic.keys())]



import functools
import operator
from optimus.engines.base.cudf.rows import CUDFBaseRows

import cudf
import pandas as pd
from optimus.engines.base.dataframe.rows import DataFrameBaseRows
from optimus.engines.base.cudf.rows import CUDFBaseRows
from optimus.engines.base.rows import BaseRows


class Rows(DataFrameBaseRows, CUDFBaseRows, BaseRows):
    pass


from optimus.helpers.types import *


class BaseEncoding:
    def __init__(self, root: 'DataFrameType'):
        self.root = root

    def n_gram(self, col, n=2):
        """
        Converts the input array of strings inside of DF into an array of n-grams.
        :param col: Column to analyze.
        :param n: number of elements per n-gram >=1.
        :return:
        """
        raise NotImplementedError('Not implemented yet')

    def one_hot_encoder(self, cols, prefix=None, drop=True, **kwargs):
        """
        Maps a column of label indices to a column of binary vectors, with at most a single one-value.
        :param cols: Columns to be encoded.
        :param prefix: Column where the output is going to be saved.
        :param drop:
        :return: Dataframe with encoded columns.
        """
        raise NotImplementedError('Not implemented yet')

    def vector_assembler(self, cols, output_col=None):
        """
        Combines a given list of columns into a single vector column.
        :param cols: Columns to be assembled.
        :param output_col: Column where the output is going to be saved.
        :return: Dataframe with assembled column.
        """
        raise NotImplementedError('Not implemented yet')

    def normalizer(df, cols, output_col=None, p=2.0):
        """
        Transforms a dataset of Vector rows, normalizing each Vector to have unit norm. It takes parameter p, which
        specifies the p-norm used for normalization. (p=2) by default.
        :param cols: Columns to be normalized.
        :param output_col: Column where the output is going to be saved.
        :param p:  p-norm used for normalization.
        :return: Dataframe with normalized columns.
        """
        raise NotImplementedError('Not implemented yet')



def unzip(val):
    list_of_tuples = list(zip(*val))
    return [list(t) for t in list_of_tuples]

def val_to_list(val, allow_none=False, convert_tuple=False):
    """
    Convert a single value to a list
    :param val: Value to convert to list
    :param allow_none: Convert the value even if it's None
    :param convert_tuple: Convert to list if it's a tuple
    :return:
    """
    if val is not None or allow_none:
        if not isinstance(val, (list, tuple) if convert_tuple else (list,)):
            val = [val]
        if isinstance(val, tuple):
            val = list(val)
    return val


def one_list_to_val(val, convert_tuple=False):
    """
    Convert a single list element to val
    :param val:
    :return:
    """
    if isinstance(val, (list, tuple) if convert_tuple else (list,)) and len(val) == 1:
        result = val[0]
    else:
        result = val

    return result


def one_tuple_to_val(val):
    """
    Convert a single tuple element to val
    :param val:
    :return:
    """
    if isinstance(val, tuple) and len(val) == 1:
        result = val[0]
    else:
        result = val

    return result



from optimus.engines.base.dataframe.dataframe import DataFrameBaseDataFrame
from optimus.engines.base.pandas.dataframe import PandasBaseDataFrame
from optimus.engines.pandas.io.save import Save


class PandasDataFrame(PandasBaseDataFrame, DataFrameBaseDataFrame):

    def _assign(self, kw_columns: dict):
        kw_columns = {str(key): kw_column for key, kw_column in kw_columns.items()}
        return self.root.data.assign(**kw_columns)

    def _base_to_dfd(self, pdf, n_partitions):
        pass

    @property
    def rows(self):
        from optimus.engines.pandas.rows import Rows
        return Rows(self)

    @property
    def cols(self):
        from optimus.engines.pandas.columns import Cols
        return Cols(self)

    @property
    def save(self):
        return Save(self)

    @property
    def functions(self):
        from optimus.engines.pandas.functions import PandasFunctions
        return PandasFunctions(self)

    @property
    def mask(self):
        from optimus.engines.pandas.mask import PandasMask
        return PandasMask(self)

    @property
    def ml(self):
        from optimus.engines.pandas.ml.models import ML
        return ML(self)

    @property
    def constants(self):
        from optimus.engines.pandas.constants import Constants
        return Constants()

    @property
    def encoding(self):
        from optimus.engines.pandas.ml.encoding import Encoding
        return Encoding(self)

    def _iloc(self, lower_bound, upper_bound, copy=True):
        dfd = self.data[lower_bound: upper_bound]
        if copy:
            dfd = dfd.reset_index(drop=True)
        return self.root.new(dfd, meta=self.root.meta)

    def to_optimus_pandas(self):
        return self.root

    def to_pandas(self):
        return self.root.data



from optimus.infer import is_numeric
from optimus.helpers.columns import parse_columns, name_col
from optimus.helpers.core import one_list_to_val
from optimus.outliers.abstract_outliers_threshold import AbstractOutlierThreshold


class ZScore(AbstractOutlierThreshold):
    """
    Handle outliers using z Score
    """

    def __init__(self, df, col_name, threshold):
        """

        :para df:
        :param col_name:
        :param threshold:
        """
        if not is_numeric(threshold):
            raise TypeError("Numeric expected")

        self.df = df
        self.threshold = threshold
        self.col_name = one_list_to_val(parse_columns(df, col_name))
        self.tmp_col = name_col(col_name, "z_score")
        self.z_score = self.df[self.col_name].cols.z_score()
        # print("self.df_score",self.df_score)
        # print("self.df",self.df)
        super().__init__()

    def info(self, output="dict"):
        self.tmp_col = name_col(self.col_name, "z_score")

        # df = self.z_score()
        df = self.df
        max_z_score = df.cols.max()

        return {"count_outliers": self.count(), "count_non_outliers": self.non_outliers_count(),
                "max_z_score": max_z_score}



from optimus.helpers.filters import dict_filter
from optimus.helpers.json import dump_json
from optimus.outliers.abstract_outliers_bounds import AbstractOutlierBounds


class Tukey(AbstractOutlierBounds):
    """
    Handle outliers using inter quartile range
    """

    def __init__(self, df, col_name):
        """

        :param df: Spark Dataframe
        :param col_name: column name
        """
        self.df = df
        self.col_name = col_name

        self.lower_bound, self.upper_bound, self.q1, self.median, self.q3, self.iqr = dict_filter(
            self.whiskers(), ["lower_bound", "upper_bound", "q1", "median", "q3", "iqr"]
        )
        # print(self.lower_bound, self.upper_bound)
        super().__init__(df, col_name, self.lower_bound, self.upper_bound)

    def whiskers(self):
        """
        Get the whiskers and IQR
        :return:
        """
        iqr = self.df.cols.iqr(self.col_name, more=True)

        lower_bound = iqr["q1"] - (iqr["iqr"] * 1.5)
        upper_bound = iqr["q3"] + (iqr["iqr"] * 1.5)

        result = {"lower_bound": lower_bound, "upper_bound": upper_bound, "q1": iqr["q1"], "median": iqr["q2"],
                  "q3": iqr["q3"], "iqr": iqr["iqr"]}

        return result

    def info(self, output: str = "dict"):
        """
        Get whiskers, iqrs and outliers and non outliers count
        :return:
        """

        lower_bound = self.lower_bound
        upper_bound = self.upper_bound

        q1 = self.q1
        median = self.median
        q3 = self.q3
        iqr = self.iqr

        result = {"count_outliers": self.count(), "count_non_outliers": self.non_outliers_count(),
                  "lower_bound": lower_bound, "lower_bound_count": self.count_lower_bound(lower_bound),
                  "upper_bound": upper_bound, "upper_bound_count": self.count_upper_bound(upper_bound),
                  "q1": q1, "median": median, "q3": q3, "iqr": iqr}

        if output == "json":
            result = dump_json(result)
        return result



from optimus.engines.base.cudf.dataframe import CUDFBaseDataFrame
from optimus.engines.base.dataframe.dataframe import DataFrameBaseDataFrame
from optimus.engines.cudf.io.save import Save
from optimus.engines.pandas.dataframe import PandasDataFrame


class CUDFDataFrame(CUDFBaseDataFrame, DataFrameBaseDataFrame):

    def _base_to_dfd(self, pdf, n_partitions):
        pass

    def to_optimus_pandas(self):
        return PandasDataFrame(self.root.data.to_pandas(), op=self.op)

    @property
    def rows(self):
        from optimus.engines.cudf.rows import Rows
        return Rows(self)

    @property
    def cols(self):
        from optimus.engines.cudf.columns import Cols
        return Cols(self)

    @property
    def save(self):
        return Save(self)

    @property
    def mask(self):
        from optimus.engines.cudf.mask import CUDFMask
        return CUDFMask(self)

    @property
    def functions(self):
        from optimus.engines.cudf.functions import CUDFFunctions
        return CUDFFunctions(self)

    @property
    def constants(self):
        from optimus.engines.cudf.constants import Constants
        return Constants()

    def _iloc(self, lower_bound, upper_bound, copy=True):
        dfd = self.data[lower_bound: upper_bound]
        if copy:
            dfd = dfd.reset_index(drop=True)
        return self.root.new(dfd, meta=self.root.meta)

    def to_pandas(self):
        return self.data.to_pandas().reset_index(drop=True)

    def to_dict(self, cols="*", n=10, orient="list"):

        if n == "all":
            series = self.cols.select(cols).to_pandas()
        else:
            series = self.cols.select(cols).iloc(0, n).to_pandas()

        return series.to_dict(orient)

    def repartition(self, n=None, *args, **kwargs):
        return self.root



from pyspark.sql import functions as F

from optimus.helpers.columns import name_col
from optimus.helpers.json import dump_json
from optimus.ml import keycollision
from optimus.ml.constants import FINGERPRINT_COL, LEVENSHTEIN_DISTANCE


def levenshtein_cluster(df, input_col, threshold: int = None, output: str = "dict"):
    """
    Output the levenshtein distance in json format
    :param df: Spark Dataframe
    :param input_col: Column to be processed
    :param threshold: number
    :param output: "dict" or "json"
    :return:
    """
    # Create fingerprint
    df_fingerprint = keycollision.fingerprint(df, input_col)

    # Names
    fingerprint_col = name_col(input_col, FINGERPRINT_COL)
    distance_col_name = name_col(input_col, LEVENSHTEIN_DISTANCE)
    temp_col_1 = input_col + "_LEVENSHTEIN_1"
    temp_col_2 = input_col + "_LEVENSHTEIN_2"
    count = "count"

    # Prepare the columns to calculate the cross join
    fingerprint_count = df_fingerprint.select(input_col, fingerprint_col).groupby(input_col) \
        .agg(F.first(input_col).alias(temp_col_1), F.first(fingerprint_col).alias(temp_col_2),
             F.count(input_col).alias(count)) \
        .select(temp_col_1, temp_col_2, count).collect()

    df = df_fingerprint.select(input_col, F.col(fingerprint_col).alias(temp_col_1),
                               F.col(fingerprint_col).alias(temp_col_2)).distinct()

    # Create all the combination between the string to calculate the levenshtein distance
    df = df.select(temp_col_1).crossJoin(df.select(temp_col_2)) \
        .withColumn(distance_col_name, F.levenshtein(F.col(temp_col_1), F.col(temp_col_2)))

    # Select only the string with shortest path
    distance_col = name_col(input_col, LEVENSHTEIN_DISTANCE)
    distance_r_col = input_col + "_LEVENSHTEIN_DISTANCE_R"
    temp_r = "TEMP_R"

    if threshold is None:
        where = ((F.col(distance_col) == 0) & (F.col(temp_col_1) != F.col(temp_col_2)))
    else:
        where = (F.col(distance_col) == 0) | (F.col(distance_col) > threshold)

    df_r = (df.rows.drop(where)
            .cols.replace(distance_col, 0, None, search_by="numeric")
            .groupby(temp_col_1)
            .agg(F.min(distance_col).alias(distance_r_col))
            # .cols.rename(distance_col, distance_r_col)
            .cols.rename(temp_col_1, temp_r)).repartition(1)

    df = df.join(df_r, ((df_r[temp_r] == df[temp_col_1]) & (df_r[distance_r_col] == df[distance_col]))) \
        .select(temp_col_1, distance_col, temp_col_2).repartition(1)

    # Create the clusters/lists
    df = (df.groupby(temp_col_1)
          .agg(F.collect_list(temp_col_2), F.count(temp_col_2)))

    # Replace ngram per string
    kv_dict = {}
    for row in fingerprint_count:
        _row = list(row.asDict().values())
        kv_dict[_row[1]] = {_row[0]: _row[2]}

    result = {}
    for row in df.collect():
        _row = list(row.asDict().values())
        d = {}
        for i in _row[1]:
            key = list(kv_dict[i].keys())[0]
            value = list(kv_dict[i].values())[0]
            d[key] = value
        key = list(kv_dict[_row[0]].keys())[0]
        value = list(kv_dict[_row[0]].values())[0]
        d.update({key: value})
        result[key] = d

    # Calculate count and sum
    f = {}
    for k, v in result.items():
        _sum = 0
        for x, y in v.items():
            _sum = _sum + y
        f[k] = {"similar": v, "count": len(v), "sum": _sum}

    result = f
    if output == "json":
        result = dump_json(result)
    return result



# This file need to be send to the cluster via .addPyFile to handle the pickle problem
# This is outside the optimus folder on purpose because it cause problem importing optimus when using de udf.
# This can not import any optimus file unless it's imported via addPyFile in Spark
import datetime
import math
import os
import re
from ast import literal_eval

import fastnumbers
import hidateinfer
import pandas as pd

# This function return True or False if a string can be converted to any datatype.
from optimus.helpers.constants import CURRENCIES, US_STATES_NAMES


class Trie:
    """Regex::Trie in Python. Creates a Trie out of a list of words. The trie can be exported to a Regex pattern.
    The corresponding Regex should match much faster than a simple Regex union."""

    def __init__(self):
        self.data = {}

    def add(self, word):
        ref = self.data
        for char in word:
            ref[char] = char in ref and ref[char] or {}
            ref = ref[char]
        ref[''] = 1

    def dump(self):
        return self.data

    def quote(self, char):
        return re.escape(char)

    def _pattern(self, pData):
        data = pData
        if "" in data and len(data.keys()) == 1:
            return None

        alt = []
        cc = []
        q = 0
        for char in sorted(data.keys()):
            if isinstance(data[char], dict):
                try:
                    recurse = self._pattern(data[char])
                    alt.append(self.quote(char) + recurse)
                except:
                    cc.append(self.quote(char))
            else:
                q = 1
        cconly = not len(alt) > 0

        if len(cc) > 0:
            if len(cc) == 1:
                alt.append(cc[0])
            else:
                alt.append('[' + ''.join(cc) + ']')

        if len(alt) == 1:
            result = alt[0]
        else:
            result = "(?:" + "|".join(alt) + ")"

        if q:
            if cconly:
                result += "?"
            else:
                result = "(?:%s)?" % result
        return result

    def pattern(self):
        return self._pattern(self.dump())

    def regex(self, words):

        for word in words:
            self.add(word)
        return re.compile(r"\b" + self.pattern() + r"\b", re.IGNORECASE)


def is_datetime_str(value: str):
    try:
        pdi = hidateinfer.infer([value])
        code_count = pdi.count('%')
        value_code_count = value.count('%')
        return code_count >= 2 and value_code_count < code_count and code_count >= len(value) / 7 and any(
            [c in pdi for c in ['/', '-', ':', '%b', '%B']])
    except Exception:
        return False


def str_to_date_format(value, date_format):
    # Check this https://stackoverflow.com/questions/17134716/convert-dataframe-column-type-from-string-to-datetime-dd-mm-yyyy-format
    try:
        pd.to_datetime(value).dt.strftime(date_format)
        return True
    except ValueError:
        return False


def str_to_null(value):
    value = value.lower()
    if value == "null":
        return True
    else:
        return False


def is_null(value):
    return pd.isnull(value)


regex_us_state_compiled = Trie().regex(US_STATES_NAMES)


def str_to_data_type(_value, _data_types):
    """
    Check if value can be parsed to a tuple or and list.
    Because Spark can handle tuples we will try to transform tuples to arrays
    :param _value:
    :param _data_types:
    :return:
    """
    # return True if isinstance(_value, str) else False
    try:
        if isinstance(literal_eval((_value.encode('ascii', 'ignore')).decode("utf-8")), _data_types):
            return True
    except (ValueError, SyntaxError, AttributeError):
        return False


def is_list_value(value):
    """
    Check if an object is a list
    :param value:
    :return:
    """
    return isinstance(value, list)


def is_list_str(value):
    # try:
    #     json.loads(value)
    #     return True
    # except JSONDecodeError:
    #     return False
    # ^ \[. *\]$
    return str_to_data_type(value, (list, tuple))


def is_object_str(value):
    # return False
    return str_to_data_type(value, (dict, set))


regex_str = r"."

regex_int = r"(^\d+\.[0]+$)|(^\d+$)"  # For cudf 0.14 regex_int = r"^\d+$"
regex_int_compiled = re.compile(regex_int)

regex_decimal = r"(^\d+\.\d+$)|(^\d+$)"
regex_decimal_compiled = re.compile(regex_decimal)

regex_non_int_decimal = r"^(\d+\.\d+)$"

regex_boolean = r"\bTrue\b|\bFalse\b|\btrue\b|\bfalse\b|\b0\b|\b1\b"
regex_boolean_compiled = re.compile(regex_boolean)

regex_list = r"^\[.*\]$"
regex_list_compiled = re.compile(regex_list)

regex_dict = r"^\{.*\}$"
regex_dict_compiled = re.compile(regex_dict)

regex_tuple = r"^\(.*\)$"
regex_tuple_compiled = re.compile(regex_tuple)


def is_bool_str(value, compile=False):
    return str_to(str(value), regex_boolean, regex_boolean_compiled, compile)


regex_gender = r"\bmale\b|\bfemale\b"
regex_gender_compiled = re.compile(regex_gender)


def is_gender(value, compile=False):
    return str_to(value, regex_gender, regex_gender_compiled, compile)


regex_full_url = (r'^(?:http|ftp)s?://'  # http:// or https://
                  r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
                  r'localhost|'  # localhost...
                  r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
                  r'(?::\d+)?'  # optional port
                  r'(?:/?|[/?]\S+)$')

regex_full_url_compiled = re.compile(regex_full_url, re.IGNORECASE)


def str_to_full_url(value, compile=False):
    return str_to(value, regex_url, regex_url_compiled, compile)


regex_url = r"([a-zA-Z]+\.)?([a-zA-Z0-9-]+\.)+([a-zA-Z]+)\/?"

regex_url_compiled = re.compile(regex_url, re.IGNORECASE)


def str_to_url(value, compile=False):
    return str_to(value, regex_url, regex_url_compiled, compile)


regex_ipv4_address = r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}"
regex_ipv4_address_compiled = re.compile(regex_ipv4_address, re.IGNORECASE)

# regex_ipv6_address = r"(?i)(?:[\da-f]{0,4}:){1,7}(?:(?<ipv4>(?:(?:25[0-5]|2[0-4]\d|1?\d\d?)\.){3}(?:25[0-5]|2[0-4]\d|1?\d\d?))|[\da-f]{0,4})"
IPV4SEG = r'(?:25[0-5]|(?:2[0-4]|1{0,1}[0-9]){0,1}[0-9])'
IPV4ADDR = r'(?:(?:' + IPV4SEG + r'\.){3,3}' + IPV4SEG + r')'
IPV6SEG = r'(?:(?:[0-9a-fA-F]){1,4})'
IPV6GROUPS = (
    r'(?:' + IPV6SEG + r':){7,7}' + IPV6SEG,  # 1:2:3:4:5:6:7:8
    r'(?:' + IPV6SEG + r':){1,7}:',  # 1::                                 1:2:3:4:5:6:7::
    r'(?:' + IPV6SEG + r':){1,6}:' + IPV6SEG,  # 1::8               1:2:3:4:5:6::8   1:2:3:4:5:6::8
    r'(?:' + IPV6SEG + r':){1,5}(?::' + IPV6SEG + r'){1,2}',  # 1::7:8             1:2:3:4:5::7:8   1:2:3:4:5::8
    r'(?:' + IPV6SEG + r':){1,4}(?::' + IPV6SEG + r'){1,3}',  # 1::6:7:8           1:2:3:4::6:7:8   1:2:3:4::8
    r'(?:' + IPV6SEG + r':){1,3}(?::' + IPV6SEG + r'){1,4}',  # 1::5:6:7:8         1:2:3::5:6:7:8   1:2:3::8
    r'(?:' + IPV6SEG + r':){1,2}(?::' + IPV6SEG + r'){1,5}',  # 1::4:5:6:7:8       1:2::4:5:6:7:8   1:2::8
    IPV6SEG + r':(?:(?::' + IPV6SEG + r'){1,6})',  # 1::3:4:5:6:7:8     1::3:4:5:6:7:8   1::8
    r':(?:(?::' + IPV6SEG + r'){1,7}|:)',  # ::2:3:4:5:6:7:8    ::2:3:4:5:6:7:8  ::8       ::
    r'fe80:(?::' + IPV6SEG + r'){0,4}%[0-9a-zA-Z]{1,}',
    # fe80::7:8%eth0     fe80::7:8%1  (link-local IPv6 addresses with zone index)
    r'::(?:ffff(?::0{1,4}){0,1}:){0,1}[^\s:]' + IPV4ADDR,
    # ::255.255.255.255  ::ffff:255.255.255.255  ::ffff:0:255.255.255.255 (IPv4-mapped IPv6 addresses and IPv4-translated addresses)
    r'(?:' + IPV6SEG + r':){1,4}:[^\s:]' + IPV4ADDR,
    # 2001:db8:3:4::192.0.2.33  64:ff9b::192.0.2.33 (IPv4-Embedded IPv6 Address)
)
regex_ipv6_address = '|'.join(['(?:{})'.format(g) for g in IPV6GROUPS[::-1]])  # Reverse rows for greedy match

regex_ipv6_address_compiled = re.compile(regex_ipv6_address, re.IGNORECASE)


def str_to_ip(value, compile=False):
    return str_to(value, regex_ipv4_address, regex_ipv4_address_compiled, compile)


# regex_email = r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)" # This do not work in CUDF/RE2
regex_email = r"^[^@]+@[^@]+\.[a-zA-Z]{2,}$"
regex_email_compiled = re.compile(regex_email, re.IGNORECASE)


def is_email(value, compile=False):
    return str_to(value, regex_email, regex_email_compiled, compile)


# Reference https://www.regular-expressions.info/creditcard.html
# https://codereview.stackexchange.com/questions/74797/credit-card-checking
regex_credit_card_number = (r'(4(?:\d{12}|\d{15})'  # Visa
                            r'|5[1-5]\d{14}'  # Mastercard
                            r'|6011\d{12}'  # Discover (incomplete?)
                            r'|7\d{15}'  # What's this?
                            r'|3[47]\d{13}'  # American Express
                            r')$')

regex_credit_card_compiled = re.compile(regex_credit_card_number)


def is_credit_card_number(value, compile=False):
    return str_to(value, regex_credit_card_number, regex_credit_card_compiled, compile)


regex_zip_code = r"^(\d{4,5}(?:[- ]\d{4})?)$"
regex_zip_code_compiled = re.compile(regex_zip_code, re.IGNORECASE)


def is_zip_code(value, compile=False):
    return str_to(value, regex_zip_code, regex_zip_code_compiled, compile)


def is_missing(value):
    return value == ""


regex_social_security_number = "^([1-9])(?!\1{2}-\1{2}-\1{4})[1-9]{2}-[1-9]{2}-[1-9]{4}"
regex_social_security_number_compiled = re.compile(regex_social_security_number, re.IGNORECASE)


def is_social_security_number(value, compile=False):
    return str_to(value, regex_social_security_number, regex_social_security_number_compiled, compile)


regex_http_code = "^[1-5][0-9][0-9]$"
regex_http_code_compiled = re.compile(regex_http_code)


def is_http_code(value, compile=False):
    return str_to(value, regex_http_code, regex_http_code_compiled, compile)


# Reference https://stackoverflow.com/questions/8634139/phone-validation-regex
regex_phone_number = r"^(\+\d{1,2}\s?)?1?\-?\.?\s?\(?\d{3}\)?[\s.-]?\d{3}[\s.-]?\d{4}$"
regex_phone_number_compiled = re.compile(regex_phone_number, re.IGNORECASE)


def is_phone_number(value, compile=False):
    # print("value",value, str_to(value, regex_phone_number, regex_phone_number_compiled, compile))
    return str_to(value, regex_phone_number, regex_phone_number_compiled, compile)


regex_BAN = r"^^[0-9]{7,14}$"
regex_BAN_compiled = re.compile(regex_BAN, re.IGNORECASE)

regex_uuid = "^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"
regex_uuid_compiled = re.compile(regex_uuid, re.IGNORECASE)

regex_mac_address = "^([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})$"
regex_mac_address_compiled = re.compile(regex_mac_address, re.IGNORECASE)

regex_driver_license = r"\b[A-Z](?:\d[- ]*){14}\b"
regex_driver_license_compiled = re.compile(regex_driver_license, re.IGNORECASE)

regex_address = r"\d+(\s+\w+){1,}\s+(?:st(?:\.|reet)?|dr(?:\.|ive)?|pl(?:\.|ace)?|ave(?:\.|nue)?|rd|road|lane|drive|way|court|plaza|square|run|parkway|point|pike|square|driveway|trace|park|terrace|blvd)"
regex_address_compiled = re.compile(regex_address, re.IGNORECASE)


def str_to(value, regex, compiled_regex, compile=False):
    value = str(value)
    if value is None:
        result = False
    else:
        if compile is True:
            regex = compiled_regex
        else:
            regex = regex

        result = bool(re.match(regex, value))
    return result


def str_to_int(value):
    return True if fastnumbers.isint(value) else False


def str_to_decimal(value):
    return True if fastnumbers.isfloat(value) else False


def str_to_str(value):
    return True if isinstance(value, str) else False


regex_currencies = "|".join(list(CURRENCIES.keys()))
regex_currencies_compiled = re.compile(regex_currencies)


def str_to_currency(value, compile=False):
    return str_to(value, regex_boolean, regex_boolean_compiled, compile)


def is_nan(value):
    """
    Check if a value is nan
    :param value:
    :return:
    """
    result = False
    if is_str(value):
        if value.lower() == "nan":
            result = True
    elif math.isnan(value):
        result = True
    return result


def is_none(value):
    """
    Check if a value is none
    :param value:
    :return:
    """
    result = False
    if is_str(value):
        if value.lower() == "none":
            result = True
    elif value is None:
        result = True
    return result


def is_same_class(class1, class2):
    """
    Check if 2 class are the same
    :param class1:
    :param class2:
    :return:
    """
    return class1 == class2


def is_(value, type_):
    """
    Check if a value is instance of a class
    :param value:
    :param type_:
    :return:
    """

    return isinstance(value, type_)


def is_type(type1, type2):
    """
    Check if a value is a specific class
    :param type1:
    :param type2:
    :return:
    """
    return type1 == type2


def is_function(value):
    """
    Check if a param is a function
    :param value: object to check for
    :return:
    """
    return hasattr(value, '__call__')


def is_empty_function(value):
    """
    Returns True if f is an empty function.
    """

    if not is_function(value):
        return False

    def empty_func():
        pass

    def empty_func_with_docstring():
        """
        Empty function with docstring.
        """
        pass

    empty_lambda = lambda: None

    empty_lambda_with_docstring = lambda: None
    empty_lambda_with_docstring.__doc__ = """Empty function with docstring."""

    def constants(f):
        """
        Return a tuple containing all the constants of a function without a docstring
        """
        return tuple(
            x
            for x in f.__code__.co_consts
            if x != f.__doc__
        )

    return (
                   value.__code__.co_code == empty_func.__code__.co_code and
                   constants(value) == constants(empty_func)
           ) or (
                   value.__code__.co_code == empty_func_with_docstring.__code__.co_code and
                   constants(value) == constants(empty_func_with_docstring)
           ) or (
                   value.__code__.co_code == empty_lambda.__code__.co_code and
                   constants(value) == constants(empty_lambda)
           ) or (
                   value.__code__.co_code == empty_lambda_with_docstring.__code__.co_code and
                   constants(value) == constants(empty_lambda_with_docstring)
           )


def is_list(value, mode=None):
    """
    Check if a string or any not string value is a python list
    :param value:
    :return:
    """
    if mode == "string":
        result = is_list_str(value)
    else:
        result = is_list_value(value)

    return result


def is_object(value):
    """
    Check if a string or any not string value is a python list
    :param value:
    :return:
    """
    return True if is_object_value(value) or is_object_str(value) else False


def is_bool(value):
    if is_bool_value(value):
        return True
    elif is_bool_value(value):
        return True
    return False


def is_list_empty(value):
    """
    Check is a list is empty
    :param value:
    :return:
    """
    return len(value) == 0


def is_dict(value):
    """
    Check if an object is a list
    :param value:
    :return:
    """
    return isinstance(value, dict)


def is_tuple(value):
    """
    Check if an object is a tuple
    :param value:
    :return:
    """
    return isinstance(value, tuple)


def is_list_or_tuple(value):
    """
    Check if an object is a list or a tuple
    :param value:
    :return:
    """
    return isinstance(value, list)


def is_list_of_int(value):
    """
    Check if an object is a list of integers
    :param value:
    :return:
    """
    return isinstance(value, list) and all(isinstance(elem, int) for elem in value)


def is_list_of_dicts(value):
    """
    Check if an object is a list of dictionaries
    :param value:
    :return:
    """
    return isinstance(value, list) and all(isinstance(elem, dict) for elem in value)


def is_list_with_dicts(value):
    """
    Check if an object is a list with some dictionaries
    :param value:
    :return:
    """
    return isinstance(value, list) and any(isinstance(elem, dict) for elem in value)


def is_list_of_float(value):
    """
    Check if an object is a list of floats
    :param value:
    :return:
    """
    return bool(value) and isinstance(value, list) and all(isinstance(elem, float) for elem in value)


def is_list_of_str_or_int(value):
    """
    Check if an object is a string or an integer
    :param value:
    :return:
    """
    return bool(value) and isinstance(value, list) and all(isinstance(elem, (int, str)) for elem in value)


def is_list_of_str_or_num(value):
    """
    Check if an object is string, integer or float
    :param value:
    :return:
    """
    return bool(value) and isinstance(value, list) and all(isinstance(elem, (str, int, float)) for elem in value)


def is_list_of_dask_dataframes(value):
    """
    Check if an object is a Spark DataFrame
    :param value:
    :return:
    """
    from dask.dataframe.core import DataFrame as DaskDataFrame
    return isinstance(value, list) and all(isinstance(elem, DaskDataFrame) for elem in value)


def is_filepath(file_path):
    """
    Check if a value ia a valid file path
    :param file_path:
    :return:
    """
    # the file is there
    if os.path.exists(file_path):
        return True
    # the file does not exists but write privileges are given
    elif os.access(os.path.dirname(file_path), os.W_OK):
        return True
    # can not write there
    else:
        return False


def is_ipv4(value):
    """
    Check if a value is valid ip
    :param value:
    :return:
    """
    return re.match(regex_ipv4_address_compiled, value)


def is_list_of_str(value):
    """
    Check if all elements in a list are strings
    :param value:
    :return:
    """
    return bool(value) and isinstance(value, list) and all(isinstance(elem, str) for elem in value)


def is_list_of_numeric(value):
    """
    Check if all elements in a list are int or float
    :param value:
    :return:
    """
    return bool(value) and isinstance(value, list) and all(isinstance(elem, (int, float)) for elem in value)


def is_list_of_list(value):
    """
    Check if all elements in a list are tuples
    :param value:
    :return:
    """

    return bool(value) and isinstance(value, list) and all(isinstance(elem, list) for elem in value)


def is_list_of_tuples(value):
    """
    Check if all elements in a list are tuples
    :param value:
    :return:
    """

    return bool(value) and isinstance(value, list) and all(isinstance(elem, tuple) for elem in value)


def is_list_of_one_element(value):
    """
    Check if a var is a single element
    :param value:
    :return:
    """
    if is_list_value(value):
        return len(value) == 1


def is_dict_of_one_element(value):
    """
    Check if a var is a single element
    :param value:
    :return:
    """
    if is_dict(value):
        return len(value) == 1


def is_one_element(value):
    """
    Check if a var is a single element
    :param value:
    :return:
    """
    return isinstance(value, (str, int, float, bool))


def is_num_or_str(value):
    """
    Check if a var is numeric(int, float) or string
    :param value:
    :return:
    """
    return isinstance(value, (int, float, str))


def is_str_or_int(value):
    """
    Check if a var is a single element
    :param value:
    :return:
    """

    return isinstance(value, (str, int))


def is_numeric(value):
    """
    Check if a var is a single element
    :param value:
    :return:
    """
    return isinstance(value, (int, float))


def is_numeric_like(value):
    """
    Check if a var is a single element
    :param value:
    :return:
    """
    return fastnumbers.isfloat(value) or fastnumbers.isintlike(value)


def is_str(value):
    """
    Check if an object is a string
    :param value:
    :return:
    """
    # Seems 20% faster than
    return isinstance(value, str)
    # return True if type("str") == "str" else False


def is_object_value(value):
    """
    Check if an object is a string
    :param value:
    :return:
    """
    return isinstance(value, object)


def is_list_of_dask_futures(value):
    """
    Check if an object is a list of strings
    :param value:
    :return:
    """
    from dask import distributed
    return bool(value) and isinstance(value, list) and all(
        isinstance(elem, distributed.client.Future) for elem in value)


def is_dask_future(value):
    """
    Check if an object is a list of strings
    :param value:
    :return:
    """
    from dask import distributed
    return isinstance(value, distributed.client.Future)


def is_float(value):
    return fastnumbers.isfloat(value, allow_nan=True)


def is_int(value):
    """
    Check if an object is an integer
    :param value:
    :return:
    """
    return isinstance(value, int) and not isinstance(value, bool)


def is_int_like(value):
    """
    Check if an object is an integer
    :param value:
    :return:
    """
    return fastnumbers.isintlike(value)


def is_float_like(value):
    """
    Check if a var is a float
    :param value:
    :return:
    """
    return fastnumbers.isfloat(value)


def is_url(value):
    regex = re.compile(
        r'^(?:http|ftp|hdfs)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)

    return re.match(regex, value)


def is_float(value):
    """
    Check if an object is an integer
    :param value:
    :return:
    """
    return isinstance(value, float)


def is_bool_value(value):
    return isinstance(value, bool)


def is_datetime(value):
    """
    Check if an object is a datetime
    :param value:
    :return:
    """
    result = False
    if isinstance(value, datetime.datetime):
        result = True
    # else:
    #     result = is_datetime_str(str(value))

    return result


def is_valid_datetime_format(value):
    try:
        now = datetime.datetime.strftime(datetime.datetime.now(tz=datetime.timezone.utc), value)
        datetime.datetime.strptime(now, value)
    except ValueError:
        return False
    except re.error:
        return False

    return True


def is_binary(value):
    """
    Check if an object is a bytearray
    :param value:
    :return:
    """
    return isinstance(value, bytearray)


def is_date(value):
    """
    Check if an object is a date
    :param value:
    :return:
    """

    return isinstance(value, datetime.date)



from abc import abstractmethod
from functools import reduce

from optimus.helpers.columns import parse_columns, name_col
from optimus.helpers.core import val_to_list
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_dict

from optimus.engines.base.columns import BaseColumns


class DataFrameBaseColumns(BaseColumns):

    def _names(self):
        return list(self.root.data.columns)

    def append(self, dfs):
        """

        :param dfs:
        :return:
        """

        dfs = val_to_list(dfs)

        df = self.root

        dfd = df.data.reset_index(drop=True)

        dfds = []

        for _df in dfs:
            if is_dict(_df):
                _dfd = dfd
                for col, value in _df.items():
                    if isinstance(value, (df.__class__,)):
                        value = value.data[value.cols.names()[0]]
                    _dfd[col] = value
                _df = _dfd[list(_df.keys())]
                del _dfd
            else:
                _df = _df.data
            dfds.append(_df)

        _dfd = self.F._engine.concat([_df.reset_index(drop=True) for _df in dfds], axis=1)

        for col in _dfd.columns:
            dfd[col] = _dfd[col]

        return self.root.new(dfd)

    def reverse(self, cols="*", output_cols=None):
        """
        Reverse values as strings
        :param cols: '*', list of columns names or a single column name.
        :return:
        """
        df = self.root
        return df.cols.apply(cols, "reverse", func_return_type=str,
                             filter_col_by_dtypes=df.constants.STRING_TYPES,
                             output_cols=output_cols, mode="vectorized")

    @staticmethod
    def astype(*args, **kwargs):
        pass

    @staticmethod
    def to_timestamp(cols, date_format=None, output_cols=None):
        pass

    def nest(self, cols="*", separator="", output_col=None, shape="string", drop=True):
        df = self.root

        dfd = df.data

        if output_col is None:
            output_col = name_col(cols)

        cols = parse_columns(df, cols)

        output_ordered_columns = df.cols.names()

        # cudfd do nor support apply or agg join for this operation
        if shape == "vector" or shape == "array":
            dfd = dfd.assign(**{output_col: dfd[cols].values.tolist()})

        elif shape == "string":
            dfds = [dfd[input_col].astype(str) for input_col in cols]
            dfd = dfd.assign(**{output_col: reduce((lambda x, y: x + separator + y), dfds)})

        if output_col not in output_ordered_columns:
            col_index = output_ordered_columns.index(cols[-1]) + 1
            output_ordered_columns[col_index:col_index] = [output_col]

        if drop is True:
            for input_col in cols:
                if input_col in output_ordered_columns and input_col != output_col:
                    output_ordered_columns.remove(input_col)

        return self.root.new(dfd).cols.select(output_ordered_columns)



# DataFrame = pd.DataFrame

import cudf

from optimus.engines.base.cudf.functions import CUDFBaseFunctions
from optimus.engines.base.dataframe.functions import DataFrameBaseFunctions


class CUDFFunctions(CUDFBaseFunctions, DataFrameBaseFunctions):

    _engine = cudf

    @staticmethod
    def to_list(_s):
        if hasattr(_s, "to_arrow"):
            return _s.to_arrow().to_pylist()
        else:
            return _s

    @staticmethod
    def dask_to_compatible(dfd):
        from optimus.helpers.converter import dask_dataframe_to_cudf
        return dask_dataframe_to_cudf(dfd)

    def count_zeros(self, series, *args):
        # Cudf can not handle null so we fill it with non zero values.
        non_zero_value = 1
        return (self.to_float(series).fillna(non_zero_value).values == 0).sum()

    def kurtosis(self, series):
        return self.to_float(series).kurt()

    def skew(self, series):
        return self.to_float(series).skew()

    def exp(self, series):
        return cudf.exp(self.to_float(series))

    def sqrt(self, series):
        return self.to_float(series).sqrt()

    def unique_values(self, series, *args):
        # Cudf can not handle null so we fill it with non zero values.
        return self.to_string(series).unique()

    # def mod(self, other):
    #     
    #     return cudf.mod(self.to_float(series), other)

    def radians(self, series):
        return cudf.radians(self.to_float(series))

    def degrees(self, series):
        return cudf.degrees(self.to_float(series))

    def ln(self, series):
        return self.to_float(series).log()

    def log(self, series, base=10):
        return cudf.log(self.to_float(series)) / cudf.log(base)

    def ceil(self, series):
        return self.to_float(series).ceil()

    def floor(self, series):
        return self.to_float(series).floor()

    def sin(self, series):
        return self.to_float(series).sin()

    def cos(self, series):
        return self.to_float(series).cos()

    def tan(self, series):
        return self.to_float(series).tan()

    def asin(self, series):
        return self.to_float(series).asin()

    def acos(self, series):
        return self.to_float(series).acos()

    def atan(self, series):
        return self.to_float(series).atan()

    def sinh(self, series):
        return 1 / 2 * (cudf.exp(series) - cudf.exp(-series))

    def cosh(self, series):
        return 1 / 2 * (cudf.exp(series) + cudf.exp(-series))

    def tanh(self, series):
        return self.sinh() / self.cosh()

    def asinh(self, series):
        return 1 / self.sinh()

    def acosh(self, series):
        return 1 / self.cosh()

    def atanh(self, series):
        return 1 / self.tanh()

    def cut(self, bins):
        raise NotImplementedError("Not implemented yet https://github.com/rapidsai/cudf/issues/5589")

    def normalize_chars(self, series):
        if not series.isnull().all():
            return self.to_string_accessor(series).normalize_characters()
        else:
            return series

    def format_date(self, series, current_format=None, output_format=None):

        # Some formats are no supported yet. https://github.com/rapidsai/cudf/issues/5991
        return cudf.to_datetime(series, format=current_format, errors="coerce").dt.strftime(output_format)

    def time_between(self, date_format=None):

        raise NotImplementedError("Not implemented yet see https://github.com/rapidsai/cudf/issues/1041")
        # return cudf.to_datetime(series).astype('str', format=date_format) - datetime.now().date()
